﻿using System.Collections.Generic;
using Data;
using Interactables;
using UnityEngine;
using Utilities;

public class GameLogic : MonoBehaviour
{
    private const float FarDistace = 1000;

    [SerializeField]
    private Material rayMaterial;

    [SerializeField]
    private Collider interactionPreventer;

    [SerializeField]
    private LineRenderer rayRendererPrefab;

    [SerializeField]
    private GameObject prisma;

    [SerializeField]
    private GameObject lightSource;

    private List<GameObject> colorTargetList;

    [SerializeField]
    private Goal goal;

    [SerializeField]
    private LightColor [] prismColors;

    private List<LightRay> lightRayList;
    private List<LineRenderer> rayRendererList;
    private List<BaseInteractable> activeInteractableList;

    private List<LightRay> prismaRayList;

    private bool isPlaying = false;

    public bool IsPlaying
    {
        get { return isPlaying; }
    }

    private void Awake()
    {
        this.colorTargetList = new List<GameObject>();
        this.colorTargetList.Add(GameObject.Find("target1"));
        this.colorTargetList.Add(GameObject.Find("target2"));
        this.colorTargetList.Add(GameObject.Find("target3"));
        this.colorTargetList.Add(GameObject.Find("target4"));
        this.colorTargetList.Add(GameObject.Find("target5"));
        this.colorTargetList.Add(GameObject.Find("target6"));
    }

    // Use this for initialization
    private void Start ()
    {
        this.Initialize();
    }

    public void Initialize()
    {
        this.isPlaying = true;
        this.interactionPreventer.enabled = false;
        this.goal.OnPuzzleSolved += Goal_OnPuzzleSolved;

        this.lightRayList = new List<LightRay>();

        BaseInteractable[] baseInteractables = GameObject.FindObjectsOfType<BaseInteractable>();
        this.activeInteractableList = new List<BaseInteractable>(baseInteractables);

        rayRendererList = new List<LineRenderer>();
        prismaRayList = new List<LightRay>();
    }

    private void Goal_OnPuzzleSolved(Goal goal)
    {
        Debug.LogWarning("Goal_OnPuzzleSolved");
        this.interactionPreventer.enabled = true;
        this.isPlaying = false;
    }

    // Update is called once per frame
    private void Update ()
    {
        if (Input.GetKey(KeyCode.Escape))
        {
            Application.Quit();
        }
        if (!this.isPlaying)
        {
            return;
        }

        this.prismaRayList.Clear();


        for (int i = 0 ; i < this.prismColors.Length ; i++)
        {
            LightColor prismColor = this.prismColors[i];
            GameObject target = this.colorTargetList[i];

            Vector3 rayVector = (target.transform.position - this.prisma.transform.position).normalized;
            Vector3 rayOrigin = this.prisma.transform.position;
            Ray ray = new Ray(rayOrigin, rayVector);
            LightRay prismaLightRay = new LightRay();
            prismaLightRay.ray = ray;
            prismaLightRay.end = ray.origin + ray.direction * FarDistace;
            prismaLightRay.lightColor = prismColor;
            prismaRayList.Add(prismaLightRay);
        }

        Logic();
	    CleanPreviousRender();
        Render();
	}

    private void Logic()
    {
        this.lightRayList.Clear();

        foreach (LightRay lightRay in prismaRayList)
        {
            Vector3 hitPoint;
            bool hitSomething = Raytrace(null, lightRay, out hitPoint);

            lightRay.end = hitSomething ? hitPoint : (lightRay.ray.origin + lightRay.ray.direction * FarDistace);
            this.lightRayList.Add(lightRay);
        }

        this.goal.TestSuccessCase(this.lightRayList);
    }

    private bool Raytrace(BaseInteractable excludedActiveInteractable, LightRay lightRay, out Vector3 hitPoint)
    {
        bool hitSomething = false;
        hitPoint = Vector3.zero;

        Vector3 closestHit = Vector3.zero;
        BaseInteractable closestActiveInteractable = null;
        float closestHitDistance = Mathf.Infinity;
        foreach (BaseInteractable activeInteractable in activeInteractableList)
        {
            if (excludedActiveInteractable != null && excludedActiveInteractable == activeInteractable)
            {
                continue;
            }
            RaycastHit hit;
            if (activeInteractable.Raycast(lightRay.ray, out hit))
            {
                float distanceToCurrentHit = Vector3.Distance(lightRay.ray.origin, hit.point);

                if (distanceToCurrentHit < closestHitDistance)
                {
                    closestHitDistance = distanceToCurrentHit;
                    closestActiveInteractable = activeInteractable;
                    closestHit = hit.point;
                }
            }
        }
        if (closestActiveInteractable != null)
        {
            hitSomething = true;
            hitPoint = closestHit;

            List<LightRay> redirectedRays = closestActiveInteractable.RetrieveDirectedRayList(lightRay, hitPoint);
            //Debug.LogWarning(closestActiveInteractable.name + " - " + redirectedRays.Count);

            foreach (LightRay redirectedLightRay in redirectedRays)
            {
                Vector3 redirectedHitPoint;
                bool redirectedHitSomething = Raytrace(closestActiveInteractable, redirectedLightRay, out redirectedHitPoint);

                redirectedLightRay.end = redirectedHitSomething ? redirectedHitPoint :
                    (redirectedLightRay.ray.origin + redirectedLightRay.ray.direction * FarDistace);
                this.lightRayList.Add(redirectedLightRay);
            }
        }
        return hitSomething;
    }

    private void CleanPreviousRender()
    {
        foreach (LineRenderer rayRenderer in rayRendererList)
        {
            DestroyImmediate(rayRenderer.gameObject);
        }
        rayRendererList.Clear();
    }

    private void Render()
    {
        CreateRayRenderer(Color.white, this.prisma.transform.position,
            this.lightSource.transform.position, 0.4f);
        foreach (LightRay lightRay in lightRayList)
        {
            CreateRayRenderer(GameUtilities.LightColorDatabase[lightRay.lightColor],
                lightRay.ray.origin, lightRay.end, 0.2f);
        }
    }

    private LineRenderer CreateRayRenderer(Color color, Vector3 start, Vector3 end, float rayWidth)
    {
        LineRenderer rayRenderer = Instantiate<LineRenderer>(rayRendererPrefab);
        rayRenderer.widthMultiplier = rayWidth;
        Material lineSegmentMaterial = Instantiate(this.rayMaterial);
        rayRenderer.material = lineSegmentMaterial;
        rayRenderer.SetPosition(0, start);
        rayRenderer.SetPosition(1, end);
        rayRenderer.startColor = color;
        rayRenderer.endColor = color;
        rayRendererList.Add(rayRenderer);
        return rayRenderer;
    }

}
